@@ -1,5 +1,5 @@
 #
-# Copyright 2017-2020 European Centre for Medium-Range Weather Forecasts (ECMWF).
+# Copyright 2017-2021 European Centre for Medium-Range Weather Forecasts (ECMWF).
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -21,12 +21,12 @@
 import datetime
 import functools
 import logging
-import typing as T  # noqa
+import typing as T
 
 import attr
-import numpy as np  # noqa
+import numpy as np
 
-from . import messages
+from . import abc, messages
 
 LOG = logging.getLogger(__name__)
 
@@ -48,12 +48,12 @@ GRIB_STEP_UNITS_TO_SECONDS = [
     1,
     900,
     1800,
-]
+]  # type: T.List[T.Union[int, None]]
 DEFAULT_EPOCH = datetime.datetime(1970, 1, 1)
 
 
-def from_grib_date_time(message, date_key='dataDate', time_key='dataTime', epoch=DEFAULT_EPOCH):
-    # type: (T.Mapping, str, str, datetime.datetime) -> int
+def from_grib_date_time(message, date_key="dataDate", time_key="dataTime", epoch=DEFAULT_EPOCH):
+    # type: (abc.Message, str, str, datetime.datetime) -> int
     """
     Return the number of seconds since the ``epoch`` from the values of the ``message`` keys,
     using datetime.total_seconds().
@@ -77,26 +77,29 @@ def from_grib_date_time(message, date_key='dataDate', time_key='dataTime', epoch
 
 
 def to_grib_date_time(
-    message, time_ns, date_key='dataDate', time_key='dataTime', epoch=DEFAULT_EPOCH
+    message, time_ns, date_key="dataDate", time_key="dataTime", epoch=DEFAULT_EPOCH
 ):
-    # type: (T.MutableMapping, np.datetime64, str, str, datetime.datetime) -> None
+    # type: (abc.MutableMessage, int, str, str, datetime.datetime) -> None
     time_s = int(time_ns) * 1e-9
     time = epoch + datetime.timedelta(seconds=time_s)
     datetime_iso = str(time)
-    message[date_key] = int(datetime_iso[:10].replace('-', ''))
-    message[time_key] = int(datetime_iso[11:16].replace(':', ''))
+    message[date_key] = int(datetime_iso[:10].replace("-", ""))
+    message[time_key] = int(datetime_iso[11:16].replace(":", ""))
 
 
-def from_grib_step(message, step_key='endStep', step_unit_key='stepUnits'):
-    # type: (T.Mapping, str, str) -> float
-    to_seconds = GRIB_STEP_UNITS_TO_SECONDS[message[step_unit_key]]
-    return message[step_key] * to_seconds / 3600.0
+def from_grib_step(message, step_key="endStep", step_unit_key="stepUnits"):
+    # type: (abc.Message, str, str) -> float
+    step_unit = message[step_unit_key]
+    to_seconds = GRIB_STEP_UNITS_TO_SECONDS[step_unit]
+    if to_seconds is None:
+        raise ValueError("unsupported stepUnit %r" % step_unit)
+    assert isinstance(to_seconds, int)  # mypy misses this
+    return int(message[step_key]) * to_seconds / 3600.0
 
 
-def to_grib_step(message, step_ns, step_unit=1, step_key='endStep', step_unit_key='stepUnits'):
-    # type: (T.MutableMapping, int, int, str, str) -> None
-    # step_seconds = np.timedelta64(step, 's').astype(int)
-    step_s = int(step_ns) * 1e-9
+def to_grib_step(message, step_ns, step_unit=1, step_key="endStep", step_unit_key="stepUnits"):
+    # type: (abc.MutableMessage, int, int, str, str) -> None
+    step_s = step_ns * 1e-9
     to_seconds = GRIB_STEP_UNITS_TO_SECONDS[step_unit]
     if to_seconds is None:
         raise ValueError("unsupported stepUnit %r" % step_unit)
@@ -104,7 +107,8 @@ def to_grib_step(message, step_ns, step_unit=1, step_key='endStep', step_unit_ke
     message[step_unit_key] = step_unit
 
 
-def from_grib_month(message, verifying_month_key='verifyingMonth', epoch=DEFAULT_EPOCH):
+def from_grib_month(message, verifying_month_key="verifyingMonth", epoch=DEFAULT_EPOCH):
+    # type: (abc.Message, str, datetime.datetime) -> int
     date = message[verifying_month_key]
     year = date // 100
     month = date % 100
@@ -112,6 +116,11 @@ def from_grib_month(message, verifying_month_key='verifyingMonth', epoch=DEFAULT
     return int((data_datetime - epoch).total_seconds())
 
 
+def to_grib_dummy(message, value):
+    # type: (abc.MutableMessage, T.Any) -> None
+    pass
+
+
 def build_valid_time(time, step):
     # type: (np.ndarray, np.ndarray) -> T.Tuple[T.Tuple[str, ...], np.ndarray]
     """
@@ -127,31 +136,31 @@ def build_valid_time(time, step):
         dims = ()  # type: T.Tuple[str, ...]
     elif len(time.shape) > 0 and len(step.shape) == 0:
         data = time + step_s
-        dims = ('time',)
+        dims = ("time",)
     elif len(time.shape) == 0 and len(step.shape) > 0:
         data = time + step_s
-        dims = ('step',)
+        dims = ("step",)
     else:
         data = time[:, None] + step_s[None, :]
-        dims = ('time', 'step')
+        dims = ("time", "step")
     return dims, data
 
 
 COMPUTED_KEYS = {
-    'time': (from_grib_date_time, to_grib_date_time),
-    'step': (from_grib_step, to_grib_step),
-    'valid_time': (
-        functools.partial(from_grib_date_time, date_key='validityDate', time_key='validityTime'),
-        functools.partial(to_grib_date_time, date_key='validityDate', time_key='validityTime'),
+    "time": (from_grib_date_time, to_grib_date_time),
+    "step": (from_grib_step, to_grib_step),
+    "valid_time": (
+        functools.partial(from_grib_date_time, date_key="validityDate", time_key="validityTime"),
+        functools.partial(to_grib_date_time, date_key="validityDate", time_key="validityTime"),
     ),
-    'verifying_time': (from_grib_month, None),
-    'indexing_time': (
-        functools.partial(from_grib_date_time, date_key='indexingDate', time_key='indexingTime'),
-        functools.partial(to_grib_date_time, date_key='indexingDate', time_key='indexingTime'),
+    "verifying_time": (from_grib_month, to_grib_dummy),
+    "indexing_time": (
+        functools.partial(from_grib_date_time, date_key="indexingDate", time_key="indexingTime"),
+        functools.partial(to_grib_date_time, date_key="indexingDate", time_key="indexingTime"),
     ),
-}
+}  # type: messages.ComputedKeysType
 
 
-@attr.attrs()
+@attr.attrs(auto_attribs=True)
 class CfMessage(messages.ComputedKeysMessage):
-    computed_keys = attr.attrib(default=COMPUTED_KEYS)
+    computed_keys: messages.ComputedKeysType = COMPUTED_KEYS
