@@ -1,5 +1,5 @@
 #
-# Copyright 2017-2020 European Centre for Medium-Range Weather Forecasts (ECMWF).
+# Copyright 2017-2021 European Centre for Medium-Range Weather Forecasts (ECMWF).
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,31 +17,18 @@
 #   Alessandro Amici - B-Open - https://bopen.eu
 #
 
-import collections
 import contextlib
 import hashlib
-import io
 import logging
 import os
 import pickle
 import typing as T
 
 import attr
+import eccodes  # type: ignore
 import numpy as np
 
-try:
-    # select between using the external ecCodes bindings or the internal implementation
-    if int(os.environ.get('CFGRIB_USE_EXTERNAL_ECCODES_BINDINGS', '0')):
-        import eccodes
-    else:
-        from . import bindings as eccodes
-except RuntimeError as exc:
-    # hide the pyeccodes import error from the majority of the users
-    # that have problems with the ecCodes bindings
-    try:
-        import pyeccodes.compat as eccodes
-    except ImportError:
-        raise exc
+from . import abc
 
 eccodes_version = eccodes.codes_get_api_version()
 
@@ -55,7 +42,7 @@ eccodes.codes_grib_multi_support_off()
 
 
 @contextlib.contextmanager
-def multi_enabled(file):
+def multi_enabled(file: T.IO[bytes]) -> T.Iterator[None]:
     """Context manager that enables MULTI-FIELD support in ecCodes from a clean state"""
     eccodes.codes_grib_multi_support_on()
     #
@@ -72,19 +59,31 @@ def multi_enabled(file):
     eccodes.codes_grib_multi_support_off()
 
 
-@attr.attrs()
-class Message(collections.abc.MutableMapping):
+KEY_TYPES = {
+    "float": float,
+    "int": int,
+    "str": str,
+    "": None,
+}
+
+
+OffsetType = T.Union[int, T.Tuple[int, int]]
+OffsetsType = T.List[T.Tuple[T.Tuple[T.Any, ...], T.List[OffsetType]]]
+
+
+@attr.attrs(auto_attribs=True)
+class Message(abc.MutableMessage):
     """Dictionary-line interface to access Message headers."""
 
-    codes_id = attr.attrib()
-    encoding = attr.attrib(default='ascii', type=str)
-    errors = attr.attrib(
-        default='warn', validator=attr.validators.in_(['ignore', 'warn', 'raise'])
+    codes_id: int
+    encoding: str = "ascii"
+    errors: str = attr.attrib(
+        default="warn", validator=attr.validators.in_(["ignore", "warn", "raise"])
     )
 
     @classmethod
     def from_file(cls, file, offset=None, **kwargs):
-        # type: (T.IO[bytes], int, T.Any) -> Message
+        # type: (T.IO[bytes], T.Optional[OffsetType], T.Any) -> Message
         field_in_message = 0
         if isinstance(offset, tuple):
             offset, field_in_message = offset
@@ -105,24 +104,26 @@ class Message(collections.abc.MutableMapping):
 
     @classmethod
     def from_sample_name(cls, sample_name, **kwargs):
+        # type: (str, T.Any) -> Message
         codes_id = eccodes.codes_new_from_samples(sample_name, eccodes.CODES_PRODUCT_GRIB)
         return cls(codes_id=codes_id, **kwargs)
 
     @classmethod
     def from_message(cls, message, **kwargs):
+        # type: (Message, T.Any) -> Message
         codes_id = eccodes.codes_clone(message.codes_id)
         return cls(codes_id=codes_id, **kwargs)
 
-    def __del__(self):
+    def __del__(self) -> None:
         eccodes.codes_release(self.codes_id)
 
     def message_get(self, item, key_type=None, default=_MARKER):
-        # type: (str, type, T.Any) -> T.Any
+        # type: (str, T.Optional[type], T.Any) -> T.Any
         """Get value of a given key as its native or specified type."""
         try:
             values = eccodes.codes_get_array(self.codes_id, item, key_type)
             if values is None:
-                values = ['unsupported_key_type']
+                values = ["unsupported_key_type"]
         except eccodes.KeyValueNotFoundError:
             if default is _MARKER:
                 raise KeyError(item)
@@ -134,40 +135,33 @@ class Message(collections.abc.MutableMapping):
             return values[0]
         return values
 
-    def message_set(self, item, value):
-        # type: (str, T.Any) -> None
-        set_array = isinstance(value, T.Sequence) and not isinstance(value, (str, bytes))
-        if set_array:
+    def message_set(self, item: str, value: T.Any) -> None:
+        arr = isinstance(value, (np.ndarray, T.Sequence)) and not isinstance(value, str)
+        if arr:
             eccodes.codes_set_array(self.codes_id, item, value)
         else:
             eccodes.codes_set(self.codes_id, item, value)
 
-    def message_grib_keys(self, namespace=None):
-        # type: (str) -> T.Generator[str, None, None]
+    def message_grib_keys(self, namespace: T.Optional[str] = None) -> T.Iterator[str]:
         iterator = eccodes.codes_keys_iterator_new(self.codes_id, namespace=namespace)
         while eccodes.codes_keys_iterator_next(iterator):
             yield eccodes.codes_keys_iterator_get_name(iterator)
         eccodes.codes_keys_iterator_delete(iterator)
 
-    def message_bufr_keys(self):
-        # type: () -> T.Generator[str, None, None]
-        iterator = eccodes.codes_bufr_keys_iterator_new(self.codes_id)
-        while eccodes.codes_bufr_keys_iterator_next(iterator):
-            yield eccodes.codes_bufr_keys_iterator_get_name(iterator)
-        eccodes.codes_bufr_keys_iterator_delete(iterator)
+    def __getitem__(self, item: str) -> T.Any:
+        key, _, key_type_text = item.partition(":")
+        if key_type_text not in KEY_TYPES:
+            raise ValueError("key type not supported %r" % key_type_text)
+        key_type = KEY_TYPES[key_type_text]
+        return self.message_get(key, key_type=key_type)
 
-    def __getitem__(self, item):
-        # type: (str) -> T.Any
-        return self.message_get(item)
-
-    def __setitem__(self, item, value):
-        # type: (str, T.Any) -> None
+    def __setitem__(self, item: str, value: T.Any) -> None:
         try:
             return self.message_set(item, value)
         except eccodes.GribInternalError as ex:
-            if self.errors == 'ignore':
+            if self.errors == "ignore":
                 pass
-            elif self.errors == 'raise':
+            elif self.errors == "raise":
                 raise KeyError("failed to set key %r to %r" % (item, value))
             else:
                 if isinstance(ex, eccodes.ReadOnlyError):
@@ -176,39 +170,39 @@ class Message(collections.abc.MutableMapping):
                 else:
                     LOG.warning("failed to set key %r to %r", item, value)
 
-    def __delitem__(self, item):
+    def __delitem__(self, item: str) -> None:
         raise NotImplementedError
 
-    def __iter__(self):
-        # type: () -> T.Generator[str, None, None]
+    def __iter__(self) -> T.Iterator[str]:
         for key in self.message_grib_keys():
             yield key
 
-    def __len__(self):
-        # type: () -> int
+    def __len__(self) -> int:
         return sum(1 for _ in self)
 
-    def write(self, file):
+    def write(self, file: T.IO[bytes]) -> None:
         eccodes.codes_write(self.codes_id, file)
 
 
-@attr.attrs()
+GetterType = T.Callable[..., T.Any]
+SetterType = T.Callable[..., None]
+ComputedKeysType = T.Dict[str, T.Tuple[GetterType, SetterType]]
+
+
+@attr.attrs(auto_attribs=True)
 class ComputedKeysMessage(Message):
     """Extension of Message class for adding computed keys."""
 
-    computed_keys = attr.attrib(
-        default={},
-        type=T.Dict[str, T.Tuple[T.Callable[[Message], T.Any], T.Callable[[Message], T.Any]]],
-    )
+    computed_keys: ComputedKeysType = {}
 
-    def __getitem__(self, item):
+    def __getitem__(self, item: str) -> T.Any:
         if item in self.computed_keys:
             getter, _ = self.computed_keys[item]
             return getter(self)
         else:
             return super(ComputedKeysMessage, self).__getitem__(item)
 
-    def __iter__(self):
+    def __iter__(self) -> T.Iterator[str]:
         seen = set()
         for key in super(ComputedKeysMessage, self).__iter__():
             yield key
@@ -217,7 +211,7 @@ class ComputedKeysMessage(Message):
             if key not in seen:
                 yield key
 
-    def __setitem__(self, item, value):
+    def __setitem__(self, item: str, value: T.Any) -> None:
         if item in self.computed_keys:
             _, setter = self.computed_keys[item]
             return setter(self, value)
@@ -225,19 +219,18 @@ class ComputedKeysMessage(Message):
             return super(ComputedKeysMessage, self).__setitem__(item, value)
 
 
-@attr.attrs()
-class FileStream(collections.abc.Iterable):
+@attr.attrs(auto_attribs=True)
+class FileStream(T.Iterable[Message]):
     """Iterator-like access to a filestream of Messages."""
 
-    path = attr.attrib(type=str)
-    message_class = attr.attrib(default=Message, type=Message, repr=False)
-    errors = attr.attrib(
-        default='warn', validator=attr.validators.in_(['ignore', 'warn', 'raise'])
+    path: str
+    message_class: T.Type[Message] = attr.attrib(default=Message, repr=False)
+    errors: str = attr.attrib(
+        default="warn", validator=attr.validators.in_(["ignore", "warn", "raise"])
     )
 
-    def __iter__(self):
-        # type: () -> T.Generator[Message, None, None]
-        with open(self.path, 'rb') as file:
+    def __iter__(self) -> T.Iterator[Message]:
+        with open(self.path, "rb") as file:
             # enable MULTI-FIELD support on sequential reads (like when building the index)
             with multi_enabled(file):
                 valid_message_found = False
@@ -250,29 +243,32 @@ class FileStream(collections.abc.Iterable):
                             raise EOFError("No valid message found in file: %r" % self.path)
                         break
                     except Exception:
-                        if self.errors == 'ignore':
+                        if self.errors == "ignore":
                             pass
-                        elif self.errors == 'raise':
+                        elif self.errors == "raise":
                             raise
                         else:
                             LOG.exception("skipping corrupted Message")
 
     def message_from_file(self, file, offset=None, **kwargs):
+        # type: (T.IO[bytes], T.Optional[OffsetType], T.Any) -> Message
         return self.message_class.from_file(file, offset, **kwargs)
 
-    def first(self):
-        # type: () -> Message
-        return next(iter(self))
+    def first(self) -> Message:
+        for message in self:
+            return message
+        raise ValueError("index has no message")
 
-    def index(self, index_keys, indexpath='{path}.{short_hash}.idx'):
-        # type: (T.List[str], str) -> FileIndex
+    def index(self, index_keys, indexpath="{path}.{short_hash}.idx"):
+        # type: (T.Sequence[str], str) -> FileIndex
         return FileIndex.from_indexpath_or_filestream(self, index_keys, indexpath)
 
 
 @contextlib.contextmanager
-def compat_create_exclusive(path, *args, **kwargs):
+def compat_create_exclusive(path):
+    # type: (str) -> T.Generator[T.IO[bytes], None, None]
     fd = os.open(path, os.O_WRONLY | os.O_CREAT | os.O_EXCL)
-    with io.open(fd, mode='wb', *args, **kwargs) as file:
+    with open(fd, mode="wb") as file:
         try:
             yield file
         except Exception:
@@ -281,26 +277,31 @@ def compat_create_exclusive(path, *args, **kwargs):
             raise
 
 
-@attr.attrs()
-class FileIndex(collections.abc.Mapping):
-    allowed_protocol_version = '1'
-    filestream = attr.attrib(type=FileStream)
-    index_keys = attr.attrib(type=T.List[str])
-    offsets = attr.attrib(repr=False, type=T.List[T.Tuple[T.Tuple[T.Any, ...], T.List[int]]])
-    filter_by_keys = attr.attrib(default={}, type=T.Dict[str, T.Any])
+ALLOWED_PROTOCOL_VERSION = "1"
+
+
+@attr.attrs(auto_attribs=True)
+class FileIndex(abc.Index[OffsetType, Message]):
+    filestream: FileStream
+    index_keys: T.List[str]
+    offsets: OffsetsType = attr.attrib(repr=False)
+    filter_by_keys: T.Dict[str, T.Any] = {}
+    index_protocol_version: str = ALLOWED_PROTOCOL_VERSION
 
     @classmethod
     def from_filestream(cls, filestream, index_keys):
-        offsets = collections.OrderedDict()
+        # type: (FileStream, T.Sequence[str]) -> FileIndex
+        offsets = {}  # type: T.Dict[T.Tuple[T.Any, ...], T.List[T.Union[int, T.Tuple[int, int]]]]
+        index_keys = list(index_keys)
         count_offsets = {}  # type: T.Dict[int, int]
-        header_values_cache = {}
+        header_values_cache = {}  # type: T.Dict[T.Tuple[T.Any, type], T.Any]
         for message in filestream:
             header_values = []
             for key in index_keys:
                 try:
                     value = message[key]
                 except:
-                    value = 'undef'
+                    value = "undef"
                 if isinstance(value, (np.ndarray, list)):
                     value = tuple(value)
                 # NOTE: the following ensures that values of the same type that evaluate equal are
@@ -308,7 +309,7 @@ class FileIndex(collections.abc.Mapping):
                 #   it also reduces the on-disk size of the index in a backward compatible way.
                 value = header_values_cache.setdefault((value, type(value)), value)
                 header_values.append(value)
-            offset = message.message_get('offset', int)
+            offset = message.message_get("offset", int)
             if offset in count_offsets:
                 count_offsets[offset] += 1
                 offset_field = (offset, count_offsets[offset])
@@ -318,25 +319,30 @@ class FileIndex(collections.abc.Mapping):
             offsets.setdefault(tuple(header_values), []).append(offset_field)
         self = cls(filestream=filestream, index_keys=index_keys, offsets=list(offsets.items()))
         # record the index protocol version in the instance so it is dumped with pickle
-        self.index_protocol_version = cls.allowed_protocol_version
         return self
 
     @classmethod
     def from_indexpath(cls, indexpath):
-        with io.open(indexpath, 'rb') as file:
-            return pickle.load(file)
+        # type: (str) -> FileIndex
+        with open(indexpath, "rb") as file:
+            index = pickle.load(file)
+            if not isinstance(index, cls):
+                raise ValueError("on-disk index not of expected type {cls}")
+            if index.index_protocol_version != ALLOWED_PROTOCOL_VERSION:
+                raise ValueError("protocol versione to allowed {index.index_protocol_version}")
+            return index
 
     @classmethod
     def from_indexpath_or_filestream(
-        cls, filestream, index_keys, indexpath='{path}.{short_hash}.idx', log=LOG
+        cls, filestream, index_keys, indexpath="{path}.{short_hash}.idx", log=LOG
     ):
-        # type: (FileStream, T.List[str], str, logging.Logger) -> FileIndex
+        # type: (FileStream, T.Sequence[str], str, logging.Logger) -> FileIndex
 
         # Reading and writing the index can be explicitly suppressed by passing indexpath==''.
         if not indexpath:
             return cls.from_filestream(filestream, index_keys)
 
-        hash = hashlib.md5(repr(index_keys).encode('utf-8')).hexdigest()
+        hash = hashlib.md5(repr(index_keys).encode("utf-8")).hexdigest()
         indexpath = indexpath.format(path=filestream.path, hash=hash, short_hash=hash[:5])
         try:
             with compat_create_exclusive(indexpath) as new_index_file:
@@ -353,11 +359,10 @@ class FileIndex(collections.abc.Mapping):
             filestream_mtime = os.path.getmtime(filestream.path)
             if index_mtime >= filestream_mtime:
                 self = cls.from_indexpath(indexpath)
-                allowed_protocol_version = self.allowed_protocol_version
                 if (
-                    getattr(self, 'index_keys', None) == index_keys
-                    and getattr(self, 'filestream', None) == filestream
-                    and getattr(self, 'index_protocol_version', None) == allowed_protocol_version
+                    getattr(self, "index_keys", None) == index_keys
+                    and getattr(self, "filestream", None) == filestream
+                    and getattr(self, "index_protocol_version", None) == ALLOWED_PROTOCOL_VERSION
                 ):
                     return self
                 else:
@@ -369,34 +374,36 @@ class FileIndex(collections.abc.Mapping):
 
         return cls.from_filestream(filestream, index_keys)
 
-    def __iter__(self):
+    def __iter__(self) -> T.Iterator[str]:
         return iter(self.index_keys)
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self.index_keys)
 
     @property
-    def header_values(self):
-        if not hasattr(self, '_header_values'):
-            self._header_values = {}
+    def header_values(self) -> T.Dict[str, T.List[T.Any]]:
+        if not hasattr(self, "_header_values"):
+            all_header_values = {}  # type: T.Dict[str, T.Dict[T.Any, None]]
             for header_values, _ in self.offsets:
                 for i, value in enumerate(header_values):
-                    values = self._header_values.setdefault(self.index_keys[i], [])
+                    values = all_header_values.setdefault(self.index_keys[i], {})
                     if value not in values:
-                        values.append(value)
+                        values[value] = None
+            self._header_values = {k: list(v) for k, v in all_header_values.items()}
         return self._header_values
 
-    def __getitem__(self, item):
-        # type: (str) -> list
+    def __getitem__(self, item: str) -> T.List[T.Any]:
         return self.header_values[item]
 
     def getone(self, item):
+        # type: (str) -> T.Any
         values = self[item]
         if len(values) != 1:
             raise ValueError("not one value for %r: %r" % (item, len(values)))
         return values[0]
 
     def subindex(self, filter_by_keys={}, **query):
+        # type: (T.Mapping[str, T.Any], T.Any) -> FileIndex
         query.update(filter_by_keys)
         raw_query = [(self.index_keys.index(k), v) for k, v in query.items()]
         offsets = []
@@ -414,7 +421,7 @@ class FileIndex(collections.abc.Mapping):
         )
         return index
 
-    def first(self):
-        with open(self.filestream.path, 'rb') as file:
+    def first(self) -> Message:
+        with open(self.filestream.path, "rb") as file:
             first_offset = self.offsets[0][1][0]
             return self.filestream.message_from_file(file, offset=first_offset)
