@@ -1,5 +1,5 @@
 #
-# Copyright 2017-2020 European Centre for Medium-Range Weather Forecasts (ECMWF).
+# Copyright 2017-2021 European Centre for Medium-Range Weather Forecasts (ECMWF).
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -18,12 +18,11 @@
 #
 
 import logging
-import typing as T  # noqa
-import warnings
+import typing as T
 
 import xarray as xr
 
-from . import DatasetBuildError, open_fileindex
+from .dataset import DatasetBuildError, open_fileindex
 
 LOGGER = logging.getLogger(__name__)
 
@@ -33,16 +32,17 @@ def open_dataset(path, **kwargs):
     """
     Return a ``xr.Dataset`` with the requested ``backend_kwargs`` from a GRIB file.
     """
-    if 'engine' in kwargs and kwargs['engine'] != 'cfgrib':
+    if "engine" in kwargs and kwargs["engine"] != "cfgrib":
         raise ValueError("only engine=='cfgrib' is supported")
-    kwargs['engine'] = 'cfgrib'
-    return xr.backends.api.open_dataset(path, **kwargs)
+    kwargs["engine"] = "cfgrib"
+    return xr.open_dataset(path, **kwargs)  # type: ignore
 
 
 def merge_datasets(datasets, **kwargs):
-    merged = []
+    # type: (T.Sequence[xr.Dataset], T.Any) -> T.List[xr.Dataset]
+    merged = []  # type: T.List[xr.Dataset]
     for ds in datasets:
-        ds.attrs.pop('history', None)
+        ds.attrs.pop("history", None)
         for i, o in enumerate(merged):
             if all(o.attrs[k] == ds.attrs[k] for k in o.attrs):
                 try:
@@ -69,47 +69,46 @@ def raw_open_datasets(path, backend_kwargs={}, **kwargs):
     #   to the user a confusing error message due to exception chaining
     for fbk in fbks:
         bks = backend_kwargs.copy()
-        bks['filter_by_keys'] = fbk
+        bks["filter_by_keys"] = fbk
         datasets.extend(raw_open_datasets(path, backend_kwargs=bks, **kwargs))
     return datasets
 
 
 def open_variable_datasets(path, backend_kwargs={}, **kwargs):
+    # type: (str, T.Dict[str, T.Any], T.Any) -> T.List[xr.Dataset]
     fileindex_kwargs = {
         key: backend_kwargs[key]
-        for key in ['filter_by_keys', 'indexpath', 'grib_errors']
+        for key in ["filter_by_keys", "indexpath", "grib_errors"]
         if key in backend_kwargs
     }
     index = open_fileindex(path, **fileindex_kwargs)
-    datasets = []
-    for param_id in sorted(index['paramId']):
+    datasets = []  # type: T.List[xr.Dataset]
+    for param_id in sorted(index["paramId"]):
         bk = backend_kwargs.copy()
-        bk['filter_by_keys'] = backend_kwargs.get('filter_by_keys', {}).copy()
-        bk['filter_by_keys']['paramId'] = param_id
+        bk["filter_by_keys"] = backend_kwargs.get("filter_by_keys", {}).copy()
+        bk["filter_by_keys"]["paramId"] = param_id
         datasets.extend(raw_open_datasets(path, bk, **kwargs))
     return datasets
 
 
-def open_datasets(path, no_warn=False, backend_kwargs={}, **kwargs):
+def open_datasets(path, backend_kwargs={}, **kwargs):
+    # type: (str, T.Dict[str, T.Any], T.Any) -> T.List[xr.Dataset]
     """
     Open a GRIB file groupping incompatible hypercubes to different datasets via simple heuristics.
     """
-    if no_warn:
-        warnings.warn("open_datasets is now public, no_warn will be removed", FutureWarning)
-
-    squeeze = backend_kwargs.get('squeeze', True)
+    squeeze = backend_kwargs.get("squeeze", True)
     backend_kwargs = backend_kwargs.copy()
-    backend_kwargs['squeeze'] = False
+    backend_kwargs["squeeze"] = False
     datasets = open_variable_datasets(path, backend_kwargs=backend_kwargs, **kwargs)
 
-    type_of_level_datasets = {}
+    type_of_level_datasets = {}  # type: T.Dict[str, T.List[xr.Dataset]]
     for ds in datasets:
         for _, da in ds.data_vars.items():
-            type_of_level = da.attrs.get('GRIB_typeOfLevel', 'undef')
+            type_of_level = da.attrs.get("GRIB_typeOfLevel", "undef")
             type_of_level_datasets.setdefault(type_of_level, []).append(ds)
 
-    merged = []
+    merged = []  # type: T.List[xr.Dataset]
     for type_of_level in sorted(type_of_level_datasets):
-        for ds in merge_datasets(type_of_level_datasets[type_of_level], join='exact'):
+        for ds in merge_datasets(type_of_level_datasets[type_of_level], join="exact"):
             merged.append(ds.squeeze() if squeeze else ds)
     return merged
